/*
 * CLCL
 *
 * fmt_rtf.c
 *
 * Copyright (C) 1996-2003 by Nakashima Tomoaki. All rights reserved.
 *		http://www.nakka.com/
 *		nakka@nakka.com
 */

/* Include Files */
#define _INC_OLE
#include <windows.h>
#undef  _INC_OLE
#include <richedit.h>

#include "..\CLCLPlugin.h"
#include "fmt_rtf_view.h"
#include "Memory.h"
#include "String.h"
#include "Message.h"

#include "resource.h"

/* Define */
#define TOOLTIP_SIZE			1024

/* Global Variables */
HINSTANCE hInst;

static HANDLE hLib;
static HICON txt_icon;
static HWND hTxtWnd;

/* Local Function Prototypes */

/*
 * DllMain - C
 */
int WINAPI DllMain(HINSTANCE hInstance, DWORD fdwReason, PVOID pvReserved)
{
	switch (fdwReason) {
	case DLL_PROCESS_ATTACH:
		hInst = hInstance;
		break;

	case DLL_PROCESS_DETACH:
		break;

	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
		break;
	}
	return TRUE;
}

/*
 * file_read_buf - t@Cǂݍ
 */
BYTE *file_read_buf(const TCHAR *path, DWORD *ret_size, TCHAR *err_str)
{
	HANDLE hFile;
	DWORD size;
	DWORD ret;
	BYTE *buf;

	// t@CJ
	hFile = CreateFile(path, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == NULL || hFile == (HANDLE)-1) {
		message_get_error(GetLastError(), err_str);
		return NULL;
	}
	size = GetFileSize(hFile, NULL);
	if (size == 0xFFFFFFFF) {
		message_get_error(GetLastError(), err_str);
		CloseHandle(hFile);
		return NULL;
	}

	buf = (BYTE *)mem_alloc(size + 1);
	if (buf == NULL) {
		message_get_error(GetLastError(), err_str);
		CloseHandle(hFile);
		return NULL;
	}

	// t@Cǂ݂
	if (ReadFile(hFile, buf, size, &ret, NULL) == FALSE) {
		message_get_error(GetLastError(), err_str);
		mem_free(&buf);
		CloseHandle(hFile);
		return NULL;
	}
	CloseHandle(hFile);

	if (ret_size != NULL) {
		*ret_size = size;
	}
	return buf;
}

/*
 * file_write_buf - t@Cɏ
 */
BOOL file_write_buf(const TCHAR *path, const BYTE *data, const DWORD size, TCHAR *err_str)
{
	HANDLE hFile;
	DWORD ret;

	// t@CJ
	hFile = CreateFile(path, GENERIC_READ | GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == NULL || hFile == (HANDLE)-1) {
		message_get_error(GetLastError(), err_str);
		return FALSE;
	}
	// t@C̏
	if (WriteFile(hFile, data, size, &ret, NULL) == FALSE) {
		message_get_error(GetLastError(), err_str);
		CloseHandle(hFile);
		return FALSE;
	}
	CloseHandle(hFile);
	return TRUE;
}

/*
 * x2c - 16i𐔒l(TCHAR)ɕϊ
 */
static TCHAR x2c(const char *str)
{
	TCHAR ret = 0;
	TCHAR num;
	int i;

	if (str == NULL) {
		return 0;
	}
	for (i = 0; i < 2 && *str != TEXT('\0'); i++, str++) {
		if (*str >= TEXT('0') && *str <= TEXT('9')) {
			num = *str - TEXT('0');
		} else if (*str >= TEXT('a') && *str <= TEXT('f')) {
			num = *str - TEXT('a') + 10;
		} else if (*str >= TEXT('A') && *str <= TEXT('F')) {
			num = *str - TEXT('A') + 10;
		} else {
			break;
		}
		ret = ret * 16 + num;
	}
	return ret;
}

/*
 * del_rtf - RTF̍폜
 */
static void del_rtf(TCHAR *buf)
{
	TCHAR *p, *r;
	TCHAR tmp[BUF_SIZE];
	int len = lstrlen(buf);
	int qt = 0, q;

	p = buf;
	while (*p != TEXT('\0')) {
		switch (*p) {
		case TEXT('\r'):
		case TEXT('\n'):
			// CRLFXLbv
			for (r = p; *r == TEXT('\r') || *r == TEXT('\n'); r++)
				;
			MoveMemory(p, r, len - (r - buf) + 1);
			len -= r - p;
			break;

		case TEXT('{'):
			qt++;
			if (qt > 1) {
				// {...} 폜
				q = 1;
				for (r = p + 1; *r != TEXT('\0'); r++) {
					if (*r == TEXT('\\') && *(r + 1) != TEXT('\0')) {
						r++;
					} else if (*r == TEXT('{')) {
						q++;
					} else if (*r == TEXT('}')) {
						q--;
						if (q == 0) {
							r++;
							break;
						}
					}
				}
				MoveMemory(p, r, len - (r - buf) + 1);
				len -= r - p;
			} else {
				MoveMemory(p, p + 1, len - (p - buf + 1) + 1);
				len--;
			}
			break;

		case TEXT('}'):
			qt--;
			MoveMemory(p, p + 1, len - (p - buf + 1) + 1);
			len--;
			break;

		case TEXT('\\'):
			r = p + 1;
			if (*r == TEXT('\\') || *r == TEXT('{') || *r == TEXT('}')) {
				// \\, {, }
				MoveMemory(p, r, len - (r - buf) + 1);
				len--;
				p++;
			} else if (*r == TEXT('\'')) {
				// char
				r++;
				*(p++) = x2c(r);
				r += 2;
				MoveMemory(p, r, len - (r - buf) + 1);
				len -= r - p;
			} else {
				for (; *r != TEXT('\0') && *r != TEXT('\\') &&
					*r != TEXT('{') && *r != TEXT('}') &&
					*r != TEXT('\r') && *r != TEXT('\n') && *r != TEXT(' '); r++)
					;
				if (r - p + 1 < BUF_SIZE) {
					lstrcpyn(tmp, p, r - p + 1);
					if (lstrcmpi(tmp, TEXT("\\tab")) == 0) {
						// TAB
						*(p++) = TEXT('\t');
					} else if (lstrcmpi(tmp, TEXT("\\par")) == 0) {
						// CRLF
						*(p++) = TEXT('\r');
						*(p++) = TEXT('\n');
					}
				}
				if (*r == TEXT(' ')) {
					r++;
				}
				MoveMemory(p, r, len - (r - buf) + 1);
				len -= r - p;
			}
			break;

		default:
			p++;
			break;
		}
	}
}

/*
 * get_format_header - `wb_̎擾
 */
__declspec(dllexport) BOOL CALLBACK get_format_header(const HWND hWnd, const int index, FORMAT_GET_INFO *fgi)
{
	switch (index) {
	case 0:
		lstrcpy(fgi->format_name, TEXT("Rich Text Format"));
		lstrcpy(fgi->func_header, TEXT("rtf_"));
		lstrcpy(fgi->comment, TEXT("Rich Text Format"));
		return TRUE;
	}
	return FALSE;
}

/*
 * rtf_show_property - vpeB\
 */
__declspec(dllexport) BOOL CALLBACK rtf_show_property(const HWND hWnd)
{
	return FALSE;
}

/*
 * rtf_initialize - 
 */
__declspec(dllexport) BOOL CALLBACK rtf_initialize(void)
{
	if (hLib == NULL) {
		hLib = LoadLibrary(TEXT("RICHED32.DLL"));
	}
	if (txt_icon == NULL) {
		txt_icon = LoadImage(hInst, MAKEINTRESOURCE(IDI_ICON_TEXT), IMAGE_ICON, 16, 16, 0);
	}
	rtfview_regist(hInst);
	return TRUE;
}

/*
 * rtf_get_icon - `p̃ACR擾
 */
__declspec(dllexport) HICON CALLBACK rtf_get_icon(const int icon_size, BOOL *free_icon)
{
	return LoadImage(hInst, MAKEINTRESOURCE(IDI_ICON_TEXT), IMAGE_ICON, icon_size, icon_size, 0);
}

/*
 * rtf_free - I
 */
__declspec(dllexport) BOOL CALLBACK rtf_free(void)
{
	if (hLib != NULL) {
		FreeLibrary(hLib);
		hLib = NULL;
	}
	if (txt_icon != NULL) {
		DestroyIcon(txt_icon);
		txt_icon = NULL;
	}
	return TRUE;
}

/*
 * rtf_initialize_item - ACȅ
 */
__declspec(dllexport) BOOL CALLBACK rtf_initialize_item(DATA_INFO *di, const BOOL set_init_data)
{
	return FALSE;
}

/*
 * rtf_copy_data - f[^̃Rs[
 */
__declspec(dllexport) HANDLE CALLBACK rtf_copy_data(const TCHAR *format_name, const HANDLE data, DWORD *ret_size)
{
	HANDLE ret;
	BYTE *from_mem, *to_mem;
	BYTE *p;

	// `FbN
	if (data == NULL || IsBadReadPtr(data, 1) == TRUE) {
		return NULL;
	}
	// TCY擾
	if ((*ret_size = GlobalSize(data)) == 0) {
		return NULL;
	}
	// Rs[bN
	if ((from_mem = GlobalLock(data)) == NULL) {
		return NULL;
	}

	for (p = from_mem; *ret_size > (DWORD)(p - from_mem) && *p != '\0'; p++)
		;
	*ret_size = p - from_mem + 1;

	// Rs[m
	if ((ret = GlobalAlloc(GHND, *ret_size)) == NULL) {
		GlobalUnlock(data);
		return NULL;
	}
	// Rs[惍bN
	if ((to_mem = GlobalLock(ret)) == NULL) {
		GlobalFree(ret);
		GlobalUnlock(data);
		return NULL;
	}

	// Rs[
	CopyMemory(to_mem, from_mem, *ret_size);
	*(to_mem + *ret_size - 1) = '\0';

	// bN
	GlobalUnlock(ret);
	GlobalUnlock(data);
	return ret;
}

/*
 * rtf_data_to_bytes - f[^oCgɕϊ
 */
__declspec(dllexport) BYTE* CALLBACK rtf_data_to_bytes(const DATA_INFO *di, DWORD *ret_size)
{
	BYTE *mem;
	BYTE *ret;

	if (di->data == NULL || (mem = GlobalLock(di->data)) == NULL) {
		return NULL;
	}
	ret = mem_alloc(di->size);
	if (ret == NULL) {
		GlobalUnlock(di->data);
		return NULL;
	}
	CopyMemory(ret, mem, di->size);
	if (ret_size != NULL) {
		*ret_size = di->size;
	}
	GlobalUnlock(di->data);
	return ret;
}

/*
 * rtf_bytes_to_data - oCgf[^ɕϊ
 */
__declspec(dllexport) HANDLE CALLBACK rtf_bytes_to_data(const TCHAR *format_name, const BYTE *data, DWORD *size)
{
	BYTE *ret;
	BYTE *mem;

	// Rs[m
	if (data == NULL || (ret = GlobalAlloc(GHND, *size)) == NULL) {
		return NULL;
	}
	// Rs[惍bN
	if ((mem = GlobalLock(ret)) == NULL) {
		GlobalFree(ret);
		return NULL;
	}
	// Rs[
	CopyMemory(mem, data, *size);
	// bN
	GlobalUnlock(ret);
	return ret;
}

/*
 * rtf_get_file_info - R_CAO̎擾
 */
__declspec(dllexport) int CALLBACK rtf_get_file_info(const TCHAR *format_name, const DATA_INFO *di, OPENFILENAME *of, const BOOL mode)
{
	of->lpstrFilter = TEXT("*.rtf\0*.rtf\0*.*\0*.*\0\0");
	of->nFilterIndex = 1;
	of->lpstrDefExt = TEXT("rtf");
	return 1;
}

/*
 * rtf_data_to_file - f[^t@Cɕۑ
 */
__declspec(dllexport) BOOL CALLBACK rtf_data_to_file(DATA_INFO *di, const TCHAR *file_name, const int filter_index, TCHAR *err_str)
{
	BYTE *tmp;

	if (di->data == NULL) {
		return FALSE;
	}
	if ((tmp = GlobalLock(di->data)) == NULL) {
		message_get_error(GetLastError(), err_str);
		return FALSE;
	}
	// t@Cɏ
	if (file_write_buf(file_name, tmp, di->size - 1, err_str) == FALSE) {
		GlobalUnlock(di->data);
		return FALSE;
	}
	GlobalUnlock(di->data);
	return TRUE;
}

/*
 * rtf_file_to_data - t@Cf[^쐬
 */
__declspec(dllexport) HANDLE CALLBACK rtf_file_to_data(const TCHAR *file_name, const TCHAR *format_name, DWORD *ret_size, TCHAR *err_str)
{
	HANDLE ret = NULL;
	BYTE *data;
	BYTE *mem;
	DWORD size;

	// t@C̓ǂݍ
	data = file_read_buf(file_name, &size, err_str);
	if (data == NULL) {
		return NULL;
	}
	// Rs[m
	if ((ret = GlobalAlloc(GHND, size + 1)) == NULL) {
		message_get_error(GetLastError(), err_str);
		mem_free(&data);
		return NULL;
	}
	// Rs[惍bN
	if ((mem = GlobalLock(ret)) == NULL) {
		message_get_error(GetLastError(), err_str);
		GlobalFree(ret);
		mem_free(&data);
		return NULL;
	}
	// Rs[
	CopyMemory(mem, data, size);
	*(mem + size) = '\0';
	if (ret_size != NULL) {
		*ret_size = size;
	}
	// bN
	GlobalUnlock(ret);
	mem_free(&data);
	return ret;
}

/*
 * rtf_free_data - f[^̉
 */
__declspec(dllexport) BOOL CALLBACK rtf_free_data(const TCHAR *format_name, HANDLE data)
{
	if (data == NULL) {
		return TRUE;
	}
	if (GlobalFree((HGLOBAL)data) != NULL) {
		return FALSE;
	}
	return TRUE;
}

/*
 * rtf_free_item - ACẻ
 */
__declspec(dllexport) BOOL CALLBACK rtf_free_item(DATA_INFO *di)
{
	return FALSE;
}

/*
 * rtf_get_menu_title - j[^Cg̎擾
 */
__declspec(dllexport) BOOL CALLBACK rtf_get_menu_title(DATA_INFO *di)
{
	TCHAR buf[BUF_SIZE];
	TCHAR tmp[BUF_SIZE];
	BYTE *mem;
	TCHAR *p, *r;

	if (di->data == NULL || (mem = GlobalLock(di->data)) == NULL) {
		return FALSE;
	}
	if (*mem == '\0') {
		GlobalUnlock(di->data);
		return TRUE;
	}

	// j[p
	p = alloc_char_to_tchar(mem);
	del_rtf(p);
	lstrcpyn(buf, p, BUF_SIZE);
	mem_free(&p);
	GlobalUnlock(di->data);

	lstrcpy(tmp, buf);
	r = buf;
	for (p = tmp; *p == TEXT(' ') || *p == TEXT('\t') || *p == TEXT('\r') || *p == TEXT('\n'); p++)
		;
	while ((r - buf) < (BUF_SIZE - 4) && *p != TEXT('\0')) {
		if (*p == TEXT('\r') || *p == TEXT('\n')) {
			lstrcpy(r, TEXT("..."));
			r += 3;
			break;
		}
		switch (*p) {
		case TEXT(' '):
		case TEXT('\t'):
			(*r++) = TEXT(' ');
			for (; *p == TEXT(' ') || *p == TEXT('\t'); p++)
				;
			break;

		case TEXT('&'):
			(*r++) = TEXT('&');
			(*r++) = *(p++);
			break;

		default:
			(*r++) = *(p++);
			break;
		}
	}
	*r = TEXT('\0');

	if (*buf != TEXT('\0')) {
		di->menu_title = alloc_copy(buf);
		di->free_title = TRUE;
	}
	return TRUE;
}

/*
 * rtf_get_menu_icon - j[pACR̎擾
 */
__declspec(dllexport) BOOL CALLBACK rtf_get_menu_icon(DATA_INFO *di, const int icon_size)
{
	di->menu_icon = txt_icon;
	di->free_icon = FALSE;
	return TRUE;
}

/*
 * rtf_get_menu_bitmap - j[prbg}bv̎擾
 */
__declspec(dllexport) BOOL CALLBACK rtf_get_menu_bitmap(DATA_INFO *di, const int width, const int height)
{
	return FALSE;
}

/*
 * rtf_get_tooltip_text - j[pc[`bveLXg
 */
__declspec(dllexport) TCHAR* CALLBACK rtf_get_tooltip_text(DATA_INFO *di)
{
	TCHAR *ret, *p;
	BYTE *mem;

	if (di->data == NULL || (mem = GlobalLock(di->data)) == NULL) {
		return NULL;
	}
	if (*mem == '\0') {
		GlobalUnlock(di->data);
		return NULL;
	}
	ret = mem_alloc(sizeof(TCHAR) * TOOLTIP_SIZE);
	if (ret == NULL) {
		GlobalUnlock(di->data);
		return NULL;
	}
	p = alloc_char_to_tchar(mem);
	del_rtf(p);
	lstrcpyn(ret, p, TOOLTIP_SIZE);
	mem_free(&p);
	GlobalUnlock(di->data);
	return ret;
}

/*
 * rtf_window_create - f[^\EBhE̍쐬
 */
__declspec(dllexport) HWND CALLBACK rtf_window_create(const HWND parent_wnd)
{
	if (hTxtWnd == NULL) {
		hTxtWnd = rtfview_create(hInst, parent_wnd, 0);
	}
	return hTxtWnd;
}

/*
 * rtf_window_destroy - f[^\EBhE̔j
 */
__declspec(dllexport) BOOL CALLBACK rtf_window_destroy(const HWND hWnd)
{
	hTxtWnd = NULL;
	return TRUE;
}

/*
 * rtf_window_show_data - f[^̕\
 */
__declspec(dllexport) BOOL CALLBACK rtf_window_show_data(const HWND hWnd, DATA_INFO *di, const BOOL lock)
{
	BYTE *mem;
#ifdef UNICODE
	TCHAR *buf;
#endif

	if (di->data == NULL) {
		SendMessage(hWnd, WM_SETTEXT, 0, (LPARAM)TEXT(""));
	} else {
		if ((mem = GlobalLock(di->data)) == NULL) {
			return FALSE;
		}
#ifdef UNICODE
		buf = alloc_char_to_tchar(mem);
		SendMessage(hWnd, WM_SETRTF, 0, (LPARAM)buf);
		mem_free(&buf);
#else
		SendMessage(hWnd, WM_SETRTF, 0, (LPARAM)mem);
#endif
		GlobalUnlock(di->data);
	}

	if (lock == TRUE) {
		SendMessage(hWnd, EM_SETOPTIONS, ECOOP_OR, ECO_READONLY);
	} else {
		SendMessage(hWnd, EM_SETOPTIONS, ECOOP_AND, ~ECO_READONLY);
	}
	SendMessage(hWnd, EM_SETMODIFY, FALSE, 0);
	return TRUE;
}

/*
 * rtf_window_save_data - f[^̕ۑ
 */
__declspec(dllexport) BOOL CALLBACK rtf_window_save_data(const HWND hWnd, DATA_INFO *di)
{
	HANDLE data;
	BYTE *to_mem;
	DWORD size;
	HLOCAL hloc;
	TCHAR *from_mem;

	if (hWnd == NULL ||
		(SendMessage(hWnd, EM_GETOPTIONS, 0, 0) & ECO_READONLY) ||
		SendMessage(hWnd, EM_GETMODIFY, 0, 0) == FALSE) {
		return FALSE;
	}
	if (di->data != NULL) {
		GlobalFree(di->data);
		di->data = NULL;
		di->size = 0;
	}

	// ݕ\Ăe̎擾
	if ((hloc = (HLOCAL)SendMessage(hWnd, WM_GETRTF, 0, (LPARAM)&size)) == NULL) {
		return FALSE;
	}
	if ((from_mem = LocalLock(hloc)) == NULL) {
		LocalFree(hloc);
		return FALSE;
	}
	// f[^̍쐬
	if ((data = GlobalAlloc(GHND, size + 1)) == NULL) {
		LocalUnlock(hloc);
		LocalFree(hloc);
		return FALSE;
	}
	if ((to_mem = GlobalLock(data)) == NULL) {
		GlobalFree(data);
		LocalUnlock(hloc);
		LocalFree(hloc);
		return FALSE;
	}
#ifdef UNICODE
	tchar_to_char(from_mem, to_mem, size);
#else
	lstrcpy(to_mem, from_mem);
#endif
	GlobalUnlock(data);
	LocalUnlock(hloc);
	LocalFree(hloc);

	// Vf[^ݒ
	di->data = data;
	di->size = size + 1;

	SendMessage(hWnd, EM_SETMODIFY, FALSE, 0);
	return TRUE;
}

/*
 * rtf_window_hide_data - f[^̔\
 */
__declspec(dllexport) BOOL CALLBACK rtf_window_hide_data(const HWND hWnd, DATA_INFO *di)
{
	SendMessage(hWnd, WM_SETTEXT, 0, (LPARAM)TEXT(""));
	return TRUE;
}
/* End of source */
